local MemoryVersion = 2;

local L = LibStub("AceLocale-3.0"):GetLocale("Short term");
local LG = LibStub("AceLocale-3.0"):GetLocale("Short term message groups");
local MAJOR = "ShortTerm";
local ShortTerm = LibStub("AceAddon-3.0"):GetAddon(MAJOR);
local LibRingCache = LibStub:GetLibrary("LibRingCache-1.0");

GlobalShortTermMemory = {
	__VERSION__ = MemoryVersion,
	[1] = {
		Data = {},
		Events = {
			BN_WHISPER = true,
			BN_CONVERSATION = true,
		},
		Length = 100,
		Name = L.MSGGROUP_BATTLENET,
	},
};
ShortTermMemory = {
	__VERSION__ = MemoryVersion;
	[1] = {
		Data = {},
		Events = {
			AFK = true,     BATTLEGROUND_LEADER = true,
			SAY = true,     BATTLEGROUND = true,
			EMOTE = true,   GUILD_OFFICER = true,
			YELL = true,    MONSTER_SAY = true,
			WHISPER = true, MONSTER_YELL = true,
			PARTY = true,   MONSTER_EMOTE = true,
			RAID = true,    MONSTER_WHISPER = true,
			GUILD = true,   MONSTER_BOSS_EMOTE = true,
			IGNORED = true, MONSTER_BOSS_WHISPER = true,
			SKILL = true,   RAID_WARNING = true,
			LOOT = true,    RAID_LEADER = true,
			MONEY = true,   BG_NEUTRAL = true,
			BG_HORDE = true,BG_ALLIANCE = true,
			CHANNEL = true, ACHIEVEMENT = true,
			DND = true,     GUILD_ACHIEVEMENT = true,
		},
		Length = 100,
		Name = L.MSGGROUP_DEFAULT,
	},
};
ShortTermMemoryPrototype = {
	--- TODO: Scopes: Guild, ServerFaction, Server
	Name = "Default",
	Scope = "CHARACTER",
	Length = 100,
	Events = {},
	Data = {},
};

local getUserData, setUserData, getCacheInfo;
do
	local userdata = {
		__general = false,
		msggroup1 = {},
	};

	function getUserData(info)
		local d = userdata;
		for k,v in ipairs(info) do
			if type(d[v]) == "table" then
				d = d[v];
			else
				break;
			end
		end
		
		return d[1];
	end

	function setUserData(info, value)
		local d = userdata;
		for k,v in ipairs(info) do
			d[v] = d[v] or {};
			d = d[v];
		end
		
		d[1] = value;
	end
	
	function getcacheinfo(info)
		local frame, settings;
		local id = getUserData(info)['ID'];
		local Scope = getUserData(info)['Scope'];
		if Scope == "GLOBAL" then
			frame = ShortTerm.GlobalFrames[id];
			settings = GlobalShortTermMemory[id];
		else
			frame = ShortTerm.Frames[id];
			settings = ShortTermMemory[id];
		end
		
		return frame, settings;
	end
end

local function shallowcopy(tbl)
	local ret = {};
	for k, v in pairs(tbl) do
		ret[k] = v;
	end
	return ret;
end

local function pclone(tbl, halt)
	local ret = {};

	for k, v in pairs(tbl) do
		if type(v) == "table" and not halt[v] then
			halt[v] = true;
			ret[k] = pclone(tbl, halt);
			halt[v] = nil;
		elseif type(v) == "table" then
			ret[k] = "RECURSION";
		else
			ret[k] = v;
		end
	end
	return ret;
end

local function clone(tbl)
	if type(tbl) ~= "table" then
		error("tclone:"..errmsg1..type(tbl));
	end
	return pclone(tbl, {});
end


ShortTerm.MemoryVersion = MemoryVersion;
ShortTerm.Defaults = {};
ShortTerm.Defaults.Prototype = ShortTermMemoryPrototype;
ShortTerm.Defaults.Global = GlobalShortTermMemory;
ShortTerm.Defaults.Character = ShortTermMemory;
local Events = {
	SYSTEM = true,         BATTLEGROUND_LEADER = true,
	SAY = true,            GUILD = true,
	EMOTE = true,          GUILD_OFFICER = true,
	YELL = true,           MONSTER_SAY = true,
	WHISPER = true,        MONSTER_YELL = true,
	PARTY = true,          MONSTER_EMOTE = true,
	RAID = true,           MONSTER_WHISPER = true,
	RAID_LEADER = true,    MONSTER_BOSS_EMOTE = true,
	RAID_WARNING = true,   MONSTER_BOSS_WHISPER = true,
	BATTLEGROUND = true,   ERRORS = true,
	AFK = true,            SKILL = true,
	DND = true,            LOOT = true,
	IGNORED = true,        MONEY = true,
	BG_HORDE = true,       OPENING = true,
	BG_ALLIANCE = true,    TRADESKILLS = true,
	BG_NEUTRAL = true,     PET_INFO = true,
	COMBAT_XP_GAIN = true, COMBAT_MISC_INFO = true,
	ACHIEVEMENT = true,    COMBAT_HONOR_GAIN = true,
	CHANNEL = true,        COMBAT_FACTION_CHANGE = true,
	GUILD_ACHIEVEMENT = true,
	BN_WHISPER = true,     BN_CONVERSATION = true,
	BN_TOAST = true,
};


--[[ OPTIONS ]]--

local optionstable = {
	type = "group",
	name = L.SHORT_TERM_MEMORY,
	childGroups = "tab",
	args = {
		__com = {
			name = L.DISCLAIMER,
			type = "description",
			order = 2,
		},
		__gui = {
			name = "Config dialog",
			order = 1,
			guiHidden = true,
			dialogHidden = true,
			type = "execute",
			func = function() InterfaceOptionsFrame_OpenToCategory(optionstablecategory); end,
		},
		__general = {
			name = L.GENERAL,
			order = 3,
			type = "group",
			args = {
				addgroup = {
					name = L.ADDGROUP,
					order = 1,
					type = "group",
					inline = true,
					disabled = true,
					args = {
						addgroupinput = {
							name = "",
							desc = L.ADDGROUP_DESC,
							type = "input",
							set = function(info, val)
								val = tonumber(val);
							end,
							get = function() return "text"; end,
						},
						addgroupbutton = {
							name = L.ADDGROUP,
							desc = L.ADDGROUP_DESC,
							type = "execute",
							func = function() return; end,
						},
					}
				},
				bnwhisperbehaviour = {
					name = L.BNWHISP_CONV,
					desc = L.BNWHISP_CONV_DESC,
					order = 2,
					type = "select",
					style = "radio",
					values = function()
						return {
							[1] = L.BNWHISP_CONV_NOCHANGE,
							[2] = L.BNWHISP_CONV_TOWHISP,
							[3] = L.BNWHISP_CONV_NICKNAME,
						};
					end,
					get = function(info, ...)
						setUserData(info, GlobalShortTermMemory.__BNWHISPER_MODE);
						return GlobalShortTermMemory.__BNWHISPER_MODE or 1;
					end,
					set = function(info, val)
						GlobalShortTermMemory.__BNWHISPER_MODE = val;
					end,
				},
				bnwhisperbehaviourdesc = {
					name = function(info)
						local s = {
							L.BNWHISP_CONV_NOCHANGE_DESC,
							L.BNWHISP_CONV_TOWHISP_DESC,
							L.BNWHISP_CONV_NICKNAME_DESC,
						}
						local i = shallowcopy(info);
						i[#i] = "bnwhisperbehaviour";
						return s[getUserData(i) or 1];
					end,
					desc = L.BNWHISP_CONV_DESC,
					order = 3,
					type = "description",
					get = function(info, index, ...)
						return GlobalShortTermMemory.__BNWHISPER_MODE or 1;
					end,
					set = function(info, index, val)
						GlobalShortTermMemory.__BNWHISPER_MODE = val;
					end,
				},				--[[
				showfragments = {
					name = L.SHOW_FRAGMENTS,
					desc = L.SHOW_FRAGMENTS_DESC,
					type = "toggle",
					order = 3,
					set = function(info,val)
						ShortTermMemorySettings.ShowFragments = val;
					end,
					get = function()
						return ShortTermMemorySettings.ShowFragments or false;
					end,
				},
				]]
			},
		},
		-- more groups go here
	},
};

local MessageGroup = {
	name = L.TYPES,
	desc = L.TYPES_DESC,
	order = 4,
	type = "group",
	--stm_scope = "GLOBAL",
	--stm_id = 1,
	args = {
		desc = {
			name = L.TYPES_DESC,
			type = "description",
			order = 1,
		},
		bllength = {
			name = L.BACKLOG_LENGTH,
			desc = L.BACKLOG_LENGTH_DESC,
			type = "range",
			order = 2,
			min = 10,
			softMax = 500,
			max = 10000,
			step = 5,
			bigStep = 20,
			width = "full",
			set = function(info, val)
				local frame, settings = getcacheinfo(info);
				
				if settings then
					settings.Length = tonumber(val);
					-- for now lets just hope the player does not do this too often.
					-- frame.cache = LibRingCache:createBlackbox(val, ShortTermMemory);
				end
			end,
			get = function(info)
				local frame, settings = getcacheinfo(info);
				
				if settings then 
					if not settings.Length then
						settings.Length = 50;
					end
					return settings.Length;
				end
			end,
		},
		--[[
		backlogtime = {
			name = L.BACKLOG_TIME,
			desc = L.BACKLOG_TIME_DESC,
			type = "range",
			order = 2,
			disabled = true,
			min = 0,
			max = 72,
			step = 1,
			bigStep = 1,
			width = "full",
			set = function(info,val)
				ShortTermMemorySettings.BacklogTimeOut = val;
			end,
			get = function() return ShortTermMemorySettings.BacklogTimeOut or 2; end,
		},
		]]
		msgtypes = {
			name = L.TYPES2,
			type = "multiselect",
			order = 3,
			width = "full",
			values = function()
				local ks = {};
				for k,v in pairs(Events) do
					ks[k] = LG[k];
				end
				return ks;
			end,
			get = function(info, index, ...)
				_G.INSPECTINFO = info;
				_G.MOREARGRUMENTS = getUserData(info);
				local frame, settings = getcacheinfo(info);
				return settings.Events[index] or false;
			end,
			set = function(info, index, val)
				local frame, settings = getcacheinfo(info);
				local action = (val and "RegisterEvent") or "UnregisterEvent";
				-- if val true, action="RegisterEvent" else "UnregisterEvent"
				settings.Events[index] = val or nil;
				-- makes sure "FALSE"-entries dont get saved and slow down loading
				for k,Event in pairs(ShortTerm.Events[index]) do
					frame[action](frame, Event);
				end
			end,
		},
	},
};

local AceConfig = LibStub("AceConfig-3.0");
local AceConfigDialog = LibStub("AceConfigDialog-3.0");

AceConfig:RegisterOptionsTable("Short term memory", function()
	for k, v in ipairs(ShortTermMemory) do
		local cpy = shallowcopy(MessageGroup);
		cpy.name = v.Name;
		setUserData({"msggroup"..k}, {ID=k, Scope=v.Scope or "CHARACTER"});
		optionstable.args["msggroup"..k] = cpy;
	end
	for k, v in ipairs(GlobalShortTermMemory) do
		local cpy = shallowcopy(MessageGroup);
		cpy.name = v.Name;
		setUserData({"gmsggroup"..k}, {ID=k, Scope=v.Scope or "GLOBAL"});
		optionstable.args["gmsggroup"..k] = cpy;
	end
	return optionstable;
end);
optionstablecategory = AceConfigDialog:AddToBlizOptions("Short term memory", nil, nil);
